#!/usr/local/bin/rexx
/* Library to normalise a CSG function */
if arg(1) = "" then do
	call write ,"Enter expression: "
	parse pull exp
end; else
	exp = arg(1)
exp = CSGNormalize(exp)
say "Normalized=" changestr("|",exp," | ")
exit

/* --- CSGNormalize --- *
 * A tree is in normal form when all INTERSECTIONS and SUBTRACTION
 * operators have left subtree which contains NO UNION operators
 * and a right which is simply a primitive. All UNION operators are
 * pushed towards the root, and all INTERSECTION and SUBTRACTION
 * operators are pushed towards the leaves. For example,
 *  (((A+B)-C) | (((D+E)+G)-F)) | H is in normal form
 * A tree can be converted to normal form by repeatedly applying
 * the following set of production rules to the tree and then its
 * subtrees:
 * 1. X-(Y|Z) -> (X-Y)-Z
 * 2. X+(Y|Z) -> (X+Y)|(X+Z)
 * 3. X-(Y+Z) -> (X-Y)|(X-Z)
 * 4. X+(Y+Z) -> (X+Y)+Z
 * 5. X-(Y-Z) -> (X-Y)|(X+Z)
 * 6. X+(Y-Z) -> (X+Y)-Z
 * 7. (X-Y)+Z -> (X+Z)-Y
 * 8. (X|Y)-Z -> (X-Z)|(Y-Z)
 * 9. (X|Y)+Z -> (X+Z)|(Y+Z)
 * X,Y, and Z here match both primitives or subtrees.
*/
CSGNormalize: procedure
	trace o
	if arg(1)="" then return ""
	rexp = CSG_ReversePolish(arg(1))
/*say "*** RP=" rexp*/
	nexp = CSG_RNormalize(rexp)
/*say "*** Norm=" nexp*/
	call CSG_PushExp nexp
/*
	nexp = CSG_MakeNormal()
say "*** MKNORM=" nexp
	oexp = CSG_Optimize(nexp)
say "*** OPT=" oexp
	return oexp
*/
return CSG_Optimize(CSG_MakeNormal())

/* --- CSG_RNormalize --- *
 * Normalize a reverse polish expression
 */
CSG_RNormalize: procedure
	parse arg exp
	do forever
		nexp = CSG_RNormalizeRule(exp)
		if nexp = exp then return nexp
		exp = nexp
	end

/* --- CSG_RNormalizeRule --- *
 * Apply one of the above rules to the reverse polish expression
 */
CSG_RNormalizeRule: procedure
	call CSG_PushExp arg(1)
	parse pull op

	if ^datatype(op,"S") then do
		right = CSG_ReadExp();	rightop = word(right,words(right))
		left  = CSG_ReadExp();	leftop  = word(left, words(left))

		if op="-" & rightop="|" then Rule=1
		else
		if op="+" & rightop="|" then Rule=2
		else
		if op="-" & rightop="+" then Rule=3
		else
		if op="+" & rightop="+" then Rule=4
		else
		if op="-" & rightop="-" then Rule=5
		else
		if op="+" & rightop="-" then Rule=6
		else
		if op="+" & leftop="-" then Rule=7
		else
		if op="-" & leftop="|" then Rule=8
		else
		if op="+" & leftop="|" then Rule=9
		else do
			right = CSG_RNormalize(right)
			left  = CSG_RNormalize(left)
			return left right op
		end

		if Rule<=6 then do	/* X op1 (Y op2 Z) */
			X = left
			call CSG_SplitExp right
			Y = left
			Z = right
		end; else do		/* (X op2 Y) op1 Z) */
			Z = right
			call CSG_SplitExp left
			X = left
			Y = right
		end

		/*
		say ">>> RULE" Rule "  X=" X "  Y=" Y "   Z=" Z
		call CSG_PushExp X; say "X=" CSG_MakeNormal2()
		call CSG_PushExp Y; say "Y=" CSG_MakeNormal2()
		call CSG_PushExp Z; say "Z=" CSG_MakeNormal2()
		*/

		/* Normalize the subtrees */
		X = CSG_RNormalize(X)
		Y = CSG_RNormalize(Y)
		Z = CSG_RNormalize(Z)
		select
			/* 1. X-(Y|Z) -> (X-Y)-Z */
			when Rule=1 then
				return CSG_RNormalize(X Y "-") Z "-"

			/* 2. X+(Y|Z) -> (X+Y)|(X+Z) */
			when Rule=2 then
				return CSG_RNormalize(X Y "+") CSG_RNormalize(X Z "+") "|"

			/* 3. X-(Y+Z) -> (X-Y)|(X-Z) */
			when Rule=3 then
				return CSG_RNormalize(X Y "-") CSG_RNormalize(X Z "-") "|"

			/* 4. X+(Y+Z) -> (X+Y)+Z */
			when Rule=4 then
				return CSG_RNormalize(X Y "+") Z "+"

			/* 5. X-(Y-Z) -> (X-Y)|(X+Z) */
			when Rule=5 then
				return CSG_RNormalize(X Y "-") CSG_RNormalize(X Z "+") "|"

			/* 6. X+(Y-Z) -> (X+Y)-Z */
			when Rule=6 then
				return CSG_RNormalize(X Y) "+" Z "-"

			/* 7. (X-Y)+Z -> (X+Z)-Y */
			when Rule=7 then
				return CSG_RNormalize(X Z) "+" Y "-"

			/* 8. (X|Y)-Z -> (X-Z)|(Y-Z) */
			when Rule=8 then
				return CSG_RNormalize(X Z "-") CSG_RNormalize(Y Z "-") "|"

			/* 9. (X|Y)+Z -> (X+Z)|(Y+Z) */
			when Rule=9 then
				return CSG_RNormalize(X Z "+") CSG_RNormalize(Y Z "+") "|"
		end
	end
return arg(1)

/* --- CSG_Optimize --- */
/* We expect an expression in normal form, so parse with OR exp's
 * Remove terms like A+A = A, A-A = remove line from or list
 */
CSG_Optimize: procedure
	parse arg exp
/*say " >>> Optimize Exp=" exp*/
	sortexp = ""

	do product=1 while exp <> ""
		outprod = ""
		parse var exp prod "|" exp
/*say; say " >>> Optimize Product=" prod*/
		/* make an array of the terms */
		op="+"
		plus = 0
		minus = 0
		do i=1 while prod <> ""
			parse var prod term prod
			if datatype(term,"S") then do
				if op="+" then do
					if term<>"@" then do
						plus=plus+1
						plus.plus = term
					end
				end; else
				if term="@" then do
/*say " +++ Remove product -@"*/
					iterate product
				end; else do
					minus=minus+1
					minus.minus = term
				end
			end; else
				op = term
		end

		/* check for duplicates A+A=A, A-A={} */
		do i=1 to plus
			id=plus.i
			do j=i+1 to plus
				if id=plus.j then do
					plus.j=""
/*say " --- Remove +"id*/
				end
			end
			do j=1 to minus
				if id=minus.j then do
/*say " --- Remove product -"id*/
					iterate product
				end
			end
		end

		/* check for duplicates -A-A=-A */
		do i=1 to minus
			id=minus.i
			do j=i+1 to minus
				if id=minus.j then do
/*say " --- Remove -"id*/
					minus.j=""
				end
			end
		end

		/* reconstruct product */
		do i=1 to plus
			if plus.i^="" then
				outprod = outprod "+"plus.i
		end
		do i=1 to minus
			if minus.i^="" then
				outprod = outprod "-"minus.i
		end
		sortexp = sortexp space(CSG_SortStr(outprod),0)
/*say " <<< Product=" space(outprod,0)*/
	end product

	sortexp = CSG_SortStr(sortexp)
/*say " >>> SortExp=" translate(sortexp,"|"," ")*/
	new = ""
	do while sortexp <> ""
		parse var sortexp exp sortexp
		if wordpos(exp,sortexp) <> 0 then
			say "MULTIPLE EXP found" exp "," sortexp
		else
			new = new exp
	end
/*say " *** OptimizedExp=" translate(strip(new),"|"," ")*/
return translate(strip(new),"|"," ")

/* --- CSG_SortStr --- */
CSG_SortStr: procedure
	parse arg str
	n = words(str)
	word. = ""
	do i=1 while str<>""
		parse var str w str
		word.i = w
	end

	/* apply bubble sort to words */
	do i=n-1 to 1 by -1
		change = 0
		do j=1 to i
			k = j+1
			if word.j > word.k then do
				/* swap the two values */
				word = word.k
				word.k = word.j
				word.j = word
				change = 1
			end
		end
		if ^change then leave
	end
	out = ""
	do i=1 to n
		out = out word.i
	end
return out

/* --- CSG_MakeNormal --- */
/* From a reverse polish expression makeit in normal form */
CSG_MakeNormal: procedure
	parse pull op
	if ^datatype(op,"S") then do
		left = CSG_MakeNormal()
		right = CSG_MakeNormal()
		return right op left
	end
return op

/* --- CSG_MakeNormal2 --- */
CSG_MakeNormal2: procedure
	parse pull op
	if ^datatype(op,"S") then do
		left = CSG_MakeNormal2()
		right = CSG_MakeNormal2()
		return "("right op left")"
	end
return op

/* --- CSG_SplitExp --- */
CSG_SplitExp: procedure expose op left right
	call CSG_PushExp arg(1)
	parse pull op
	right = CSG_ReadExp()
	left = CSG_ReadExp()
return

/* --- CSG_ReadExp --- */
CSG_ReadExp: procedure
	if queued()=0 then return ""
	parse pull op
	if ^datatype(op,"S") then do
		right = CSG_ReadExp()
		left = CSG_ReadExp()
		return left right op
	end
return op

/* --- CSG_PushExp --- */
CSG_PushExp: procedure
	parse arg exp
	do while exp <> ""
		parse var exp w exp
		push w
	end
return

/* --- CSG_ReversePolish --- */
CSG_ReversePolish: procedure
	exp = CSG_FixExp(arg(1))
	rev = ""
	call CSG_NextSymbol
	call CSG_Exp1
return strip(rev)

/* --- CSG_FixExp --- */
CSG_FixExp: procedure
	parse arg exp
	exp = space(exp,0)
	if left(exp,1)=="-" then exp = "@"exp
	else
	if left(exp,1)=="+" then exp = substr(exp,2)

	exp = changestr("(-",exp,"(@-")
	exp = changestr("(+",exp,"(")
	exp = changestr("|-",exp,"|@-")
	exp = changestr("|+",exp,"|")

	exp = changestr("(",exp," ( ")
	exp = changestr(")",exp," ) ")
	exp = changestr("+",exp," + ")
	exp = changestr("-",exp," - ")
	exp = changestr("|",exp," | ")
return space(exp)

/* --- CSG_Exp1 --- */
CSG_Exp1: procedure expose symbol exp rev
	call CSG_Exp2
	do while symbol="|"
		call CSG_NextSymbol
		call CSG_Exp2
		rev = rev "|"
	end
	return

/* --- CSG_Exp2 --- */
CSG_Exp2: procedure expose symbol exp rev
	call CSG_Exp3
	do while symbol="+" | symbol="-"
		op = symbol
		call CSG_NextSymbol
		call CSG_Exp3
		rev = rev op
	end
	return

/* --- CSG_Exp3 --- */
CSG_Exp3: procedure expose symbol exp rev
	if symbol = "(" then do
		call CSG_NextSymbol
		call CSG_Exp1
		if symbol ^= ")" then say "ERROR unbalanced parenthesis"
		call CSG_NextSymbol
	end; else
	if datatype(symbol,"S") then do
		rev = rev symbol
		call CSG_NextSymbol
	end; else do
		say "ERROR unknown character found" symbol
		exit 999
	end
return

/* --- CSG_NextSymbol --- */
CSG_NextSymbol:
	parse var exp symbol exp
return
