#!/bin/env rexx
/*
 * Program: fluka.r
 * Description: Translates the FLUKA geometry to MCNP/X Format
 * Author: Vasilis Vlachoudis 2000,4
 */

/*
signal on error
signal on syntax
*/
signal on novalue

parse arg arg

version         = "1.0c"
lastchange_date = "19/09/2006"
released_date   = "22/10/2004"
created_date    = "4/4/4"
author          = "Vasilis.Vlachoudis@cern.ch"
parse source . . prg .

if arg="" | arg="?" | arg="-?" then do
	say "syntax:	"prg "[-options] [<input> [<output>]]"
	say "desc:	Fluka preprocessor."
	say "	Translates, Rotates, Scales and writes any fluka input"
	say "	to various formats: Fixed, Free, Numbered, MCNP(X)"
	say "options:"
	say "	v	to be more verbose"
	say "	Export:"
	say "	f	export in fluka FIXED format (default)"
	say "	d	export in fluka DOUBLE format"
	say "	F	export in fluka FREE format"
	say "	m	export in MCNP format"
	say "	M	export in MCNP format allowing multibodies (RPP, BOX...)"
	say "	p	export in POVRAY ray-tracer format (http://www.povray.org)"
	say "	P	export in POVRAY ray-tracer format (overwrite include file)"
	say "	x	export in DXF format (only primitives)"
	say "	g	export in GTS format (http://gts.sf.net)"
	say "	Geometry Transform:"
	say "	b #	blackbody region for MCNP (Default: first region)"
	say "	t <exp>	transform geometry"
	say "	a #	geometry transform operation accuracy digits (Def. 10)"
	say "	r	transform range start"
	say "	R	transform range end"
	say "	D #	material density threshold for POVRAY (default: 0.01)"
	say "	Various:"
	say "	n	convert names to numbers (Default)"
	say "	N	convert numbers to names"
	say "	e	expand region expressions"
	say "	o	optimize CSG expressions"
	say "	s #	add a suffix to names of primitives/regions for transformations"
	say "	i/I	use preprocessor"
	say "		(#if/#else...) (i=enable,I=disable default)"
	say "exp: transformation expression is a string with any of the following"
	say "	rx n		rotate on X axis by n degrees"
	say "	ry n		rotate on Y axis by n degrees"
	say "	rz n		rotate on Y axis by n degrees"
	say "	t dx dy dz	translate by [dx dy dz]"
	say "	s sx [sy sz]	scale by sx on XYZ if sy,sz are not provided"
	say "			else scale on X by sx, on Y by sy and on Z by sz"
	say "	The commands are executed in the specified order"
	say "examples:"
	say "	fluka.r -f ir7.fluka ir7.inp"
	say "	fluka.r -m ir7.inp ir7.mcnp"
	say "	fluka.r -t rz -90 yxz.inp xyz.inp"
	say "	fluka.r -t t 10 0 0 rz -90 yxz.inp xyz.inp"
	say
	say "version:	"version
	say "author:		"author
	say "created:	"created_date
	say "1st release:	"released_date
	say "last change:	"lastchange_date
exit
end

/* --- Load libraries --- */
libs = "veclib.r flukalib.r csglib.r"
do i=1 to words(libs)
	call ImportLib word(libs,i)
end
drop libs

/* --- Initialize --- */
numeric digits 16

call VecInit

filein = ""
fileout = ""

/* Constants */
@NONE      =-1
@FIXED     = 0
@DOUBLE    = 1
@FREE      = 2
@MCNP      = 3
@MCNPX     = 4
@POVRAY    = 5
@DXF       = 6
@GTS       = 7
@OUTFORMAT = "None Fixed Double Free MCNP MCNPX Povray GTS"

/* --- Global variables --- */
verbose      = 0	/* Verbose mode				*/
outFormat    = @NONE	/* Output format			*/
outDigits    = 10	/* For fixed format 10 or 22		*/
blackbody    = 1	/* Black body region			*/
toNumbers    = 1	/* Convert names to numbers		*/
expandGeom   = 0	/* Expand geometry regions		*/
optimizeGeom = 0	/* Optimize CSG geometry		*/
preprocessor = 0	/* Process preprocessor directives	*/

geoFormat    = 0
inpFormat    = 0
FarAway      = 1E10
overwritePovInc = 0	/* Overwite the pov include	*/
suffix       = ""

blank     = ""
usr.      = "?" ; usr.blank      = blank
prim.     = "?"	; prim.blank     = blank
fluka.    = ""
region.   = "?"	; region.blank   = blank
lattice.  = "?"	; lattice.blank  = blank
material. = "?" ; material.blank = blank

material.@povDensity = 0.01	/* Density threshold for pov */
transform.0         = 0		/* Transformation	*/
transform.@accuracy = 0		/* Geometry accuracy	*/
transform_start     = ""
transform_stop      = ""
transform.@scaleR   = 1

usr.@BIN.0   = 0
usr.@BDX.0   = 0
usr.@TRACK.0 = 0
usr.@YIELD.0 = 0

/* --- Process Args --- */
do while arg <> ""
	parse var arg options arg
	if left(options,1)=="-" then do
		options = substr(options,2)
		do while options <> ""
			parse var options opt +1 options
			select
				when opt=="v" then verbose = 1
				when opt=="a" then do
					if options<>""
						then transform.@accuracy=options
						else parse var arg transform.@accuracy arg
				end
				when opt=="b" then do
					if options<>""
						then blackbody=options
						else parse var arg blackbody arg
				end
				when opt=="t" then do
					arg = ParseTransform(options arg)
					options = ""
				end
				when opt=="r" then
					parse var arg transform_start arg
				when opt=="R" then
					parse var arg transform_stop arg
				when opt=="f" then do
					outFormat  = @FIXED
					toNumbers = 1
					if transform.@accuracy=0 then
						transform.@accuracy = 10
				end
				when opt=="d" then do
					outFormat  = @DOUBLE
					toNumbers = 1
					if transform.@accuracy=0 then
						transform.@accuracy = 15
				end
				when opt=="F" then do
					outFormat  = @FREE
					toNumbers = 0
					if transform.@accuracy=0 then
						transform.@accuracy = 15
				end
				when opt=="m" then do
					outFormat  = @MCNP
					if transform.@accuracy=0 then
						transform.@accuracy = 10
				end
				when opt=="M" then do
					outFormat  = @MCNPX
					if transform.@accuracy=0 then
						transform.@accuracy = 15
				end
				when opt=="g" then outFormat = @GTS
				when opt=="p" then outFormat = @POVRAY
				when opt=="P" then do
					outFormat = @POVRAY
					overwritePovInc = 1
				end
				when opt=="x" then do
					outFormat = @DXF
					if transform.@accuracy=0 then
						transform.@accuracy = 10
				end
				when opt=="s" then
					parse var arg suffix arg
				when opt=="D" then do
					if options<>"" then material.@povDensity = options
				end
				when opt=="n" then toNumbers    = 1
				when opt=="N" then toNumbers    = 0
				when opt=="e" then expandGeom   = 1
				when opt=="o" then optimizeGeom = 1
				when opt=="i" then preprocessor = 1
				when opt=="I" then preprocessor = 0
				otherwise
					call lineout "<STDERR>", "Unknown option '"opt"'"
					exit 2
			end
		end
	end; else
	if filein="" then filein = options
	else
	if fileout="" then fileout = options
	else do
		call lineout "<STDERR>", "Wrong number of options" options arg
		exit 2
	end
end

/* Fix accuracy */
if transform.@accuracy=0 then transform.@accuracy = 10
if ^datatype(transform.@accuracy,"NUM") then do
	call lineout "<STDERR>","Error in accuracy" transform.@accuracy
	exit 2
end
transform.@start = transform_start
transform.@stop  = transform_stop

/* --- Temporary for povray --- */
if ^expandGeom & @OUTFORMAT==@POVRAY then do
	say "*** WARNING: POVRAY format requested without the expand and optimize"
	say "*** WARNING: Adding option: -e"
	expandGeom = 1
end

/* --- Printout options --- */
if verbose then do
	say prg "V"version "by" author
	say "Arguments:"
	say "	Accuracy           : " transform.@accuracy "(=1E-"transform.@accuracy")"
	say "	Blackbody          : " blackbody
	say "	Output format      : " word(@OUTFORMAT,outFormat+2)
	say "	Convert to numbers : " toNumbers
	say "	Expand Regions     : " expandGeom
	say "	Optimize Geometry  : " optimizeGeom
	say "	Suffix             : " suffix
	say "	Input File         : " filein
	say "	Output File        : " fileout
	if transform.0>0 then do
		say "	Transformations:"
		say "		## Type Value"
		do i=1 to transform.0
			say "0909"x||right(i,2) left(transform.i,4) transform.i.@val
		end
		say
		say "	Transformation matrix for points:"
		call MatPrint transform.@matP,"		|",2,5," |"
		say "	Transformation matrix for vectors:"
		call MatPrint transform.@matV,"		|",2,5," |"
		say "	Transformation matrix for normals:"
		call MatPrint transform.@matN,"		|",2,5," |"
	end
	say
end

/* --- Materials --- */
/* Region 1 by default is blackbody */
call FlukaInit
region.1.@mat = 1
material.1.@used = 1

/* --- Fluka To MCNP --- */
f2mfile="fluka2mcnp.dat"
call LoadMaterialAZ f2mfile
if ^result then do
	/* check if the FLUKA2MCNP var exists */
	path = getenv("FLUKA2MCNP")
	if path<>"" then call LoadMaterialAZ path"/"f2mfile
	if ^result then do
		p = lastpos("/",prg)
		path = left(prg,p)
		call LoadMaterialAZ path"/"f2mfile
		if ^result then
			call lineout "<STDERR>", "Error: cannot locate file" f2mfile
	end
end

/* --- Files --- */
if filein<>"" then do
	fin = open(filein,"r")
	if fin<0 then call FLError "Cannot open input file '"filein"'"
end; else
	fin = 0

if filein<>"" then do
	fout = open(fileout,"w")
	if fout<0 then call FLError "Cannot open output file '"fileout"'"
end; else
	fout = 1

/* Check if only preprocessing is requested */
if outFormat=@NONE & preprocessor then do
	if verbose then say "Preprocessing file"
	do forever
		line = FlukaReadLine(fin)
		if line="" then leave
		call WriteMultiLine fout,fluka.@COMMENT
		call lineout fout,line
	end
	exit
end

/* Make 2 passes
 * 1. To read the geometry
 * 2. To read the entire file apart from the geometry
 */

/* -- Pass #1 - Scan for the geometry -- */
if verbose then say "Reading the geometry"
do forever
	line = FlukaReadLine(fin)
	if line="" then leave
	call FlukaParseLine line,inpFormat
	select
		when card = "FREE"	then do
			inpFormat=1
			if outFormat=@NONE then outFormat=inpFormat
		end
		when card = "GLOBAL"	then do
			if what.5>0 then geoFormat=@FREE
			if outFormat=@NONE then outFormat=inpFormat
		end
		when card = "GEOBEGIN"	then call ReadGeometry
	end
end
if outFormat=@NONE then outFormat=inpFormat

/* Rewind files */
fluka.@LINENO = 0
fluka.@DEFINE.0 = 0
call seek fin,0
if fin ^= fgeo then call seek fgeo,0

/* -- Pass #2 - Read entire file -- */
if verbose then say "Reading the entire input file"
inpFormat = 0

if verbose then say "Writing output" word(@OUTFORMAT,outFormat+2) "file"

/* select numeric digits */
numeric digits transform.@accuracy

do forever
	line = FlukaReadLine(fin)
	if line="" then leave
	call FlukaParseLine line,inpFormat
	if outFormat<=@FREE then call WriteMultiLine fout,fluka.@COMMENT

	select
		when card=="TITLE"	then call ReadTitle
		when card=="FREE"	then inpFormat=1
		when card=="DISCARD"	then call FixDiscard
		when card=="LOW-MAT"	then call FixLowMat
		when card=="MATERIAL"	then call ReadMaterial
		when card=="COMPOUND"	then call ReadCompound
		when card=="DELTARAY"	then call FixDeltaRay

		when card=="USRTRACK"	then call FixUsrTrack
		when card=="USRBDX"	then call FixUsrBdx
		when card=="USRBIN"	then call FixUsrBin
		when card=="EVENTBIN"	then call FixUsrXXX
		when card=="USRCOLL"	then call FixUsrXXX
		when card=="USRYIELD"	then call FixUsrBdx
		when card=="EMFCUT"	then call FixEmfCut
		when card=="EMF-BIAS"	then call FixEmfBias
		when card=="WW-FACTO"	then call FixReg45
		when card=="PAIRBREM"	then call FixMat45
		when card=="TCQUENCH"	then call FixTcquench

		when card=="LAM-BIAS"	then call FixLamBias
		when card=="PART-THR"	then call FixPartThr
		when card=="SCORE"	then call FixScore
		when card=="LOW-BIAS"	then call FixLowBias
		when card=="RESNUCLE"	then call FixResNuclei
		when card=="DUMPTHEM"	then call FixDumpThem

		when card=="GLOBAL"	then call ReadGlobal
		when card=="ASSIGNMA"	then call ReadAssignMat
		when card=="BIASING"	then call ReadBiasing
		when card=="PLOTGEOM"	then call ReadPlotGeom

		when card = "GEOBEGIN"	then do
			if outFormat<=@FREE then do
				if outFormat=@FREE then
					sdum = "COMBNAME"
				else
					sdum = "COMBINAT"
				what.3 = ""
				what.4 = ""
				call lineout fout,FlukaCreateLine(outDigits)
				line = read(fin)	/* Read first line */
				line = right(ivopt,5)||right(idbg,5)||,
					substr(line,11)
				call lineout fout,line
				call WritePrimitives fout
				call WriteRegions fout
				call WriteLattices fout
				call lineout fout,FlukaFindCard(fin,"GEOEND")
			end; else
				call FlukaFindCard fin,"GEOEND"
		end

		otherwise
			call FixAll
	end
end

/* Final checks */
/* Check regions */
do r=1 to region.0
	if region.r.@mat = "?" then
		call FLError "No material has assigned to region" region.r "["r"]"
end

/* Export other formats */
select
	/* --- Write MCNP --- */
	when outFormat=@MCNP | outFormat=@MCNPX then do
		call ImportLib "mcnplib.r"
		call lineout fout,fluka.@TITLE
		call MCNPCreateSurfaces
		call MCNPWriteCells fout
		call MCNPWriteSurfaces fout
		call MCNPWriteTransformations fout
		call MCNPWriteMaterials fout
		call MCNPWriteBiasing fout
	end
	/* --- PovRay --- */
	when outFormat=@POVRAY then do
		call ImportLib "povlib.r"
		call PovHeader fout,fileout,overwritePovInc
		call PovPrimitives fout
		call PovRegions fout
	end
	/* --- DXF --- */
	when outFormat=@DXF then do
		call ImportLib "dxflib.r"
		call DXFWriteHeader fout
		call DXFPrimitives fout
	end
	/* --- GTS --- */
	when outFormat=@GTS then do
		call ImportLib "gtslib.r"
		call GTSPrimitives fout
	end
end

/* --- Close and Exit --- */
call close fin
if fin ^= fgeo then call close fgeo
call close fout
exit 0

/* --- Import Library --- */
ImportLib: procedure
	parse arg lib
	call import lib
	if result = 1 then do
		call lineout "<STDERR>", "Error loading library" lib
		exit
	end
return

/* ========================== Routines =========================== */

/* --- ParseTransform --- */
ParseTransform: procedure expose transform.
	parse arg str
	n = transform.0
	do while str<>""
		parse var str w str
		select
			when left(w,1)=="-" & ^datatype(w,"NUM") then do
				str = w str
				leave
			end
			when w=="rx" | w=="ry" | w=="rz" then do
				parse var str deg str
				if ^datatype(deg,"NUM") then do
					call lineout "<STDERR>", "Error in transformation rotation" deg
					exit 3
				end
				n=n+1
				transform.n = w
				transform.n.@val = deg
			end
			when w=="t" then do
				parse var str dx dy dz str
				if ^datatype(dx,"NUM") |,
						^datatype(dy,"NUM") |,
						^datatype(dz,"NUM")  then do
					call lineout "<STDERR>", "Error in transformation translation" dx dy dz
					exit 3
				end
				n=n+1
				transform.n = w
				transform.n.@val = dx dy dz
			end
			when w=="s" then do
				parse var str sx sy sz str
				if ^datatype(sx,"NUM") then do
					call lineout "<STDERR>", "Error in transformation scale" sx
					exit 3
				end
				if ^datatype(sy,"NUM") | ^datatype(sz,"NUM") then do
					str = sy sz str
					sy = sx
					sz = sx
				end
				n=n+1
				transform.n = w
				transform.n.@val = sx sy sz
				transform.@scaleR = transform.@scaleR * sx
			end
			otherwise
				str = w str
				leave
		end
	end
	transform.0 = n

	/* Create the transformation matrices */
	matP = MatUnary()	/* point transformation  */
	matV = matP		/* vector transformation */
	matN = matP		/* normal transformation */
	do n=transform.0 to 1 by -1
		select
			when transform.n == "t" then
				matP = MatMult(matP,MatTrans(transform.n.@val))
			when transform.n == "s" then do
				m = MatScale(transform.n.@val)
				matP = MatMult(matP,m)
				matV = MatMult(matV,m)
			end
			otherwise
				axis = right(transform.n,1)
				/* pi/180 */
				m = MatRot(axis,transform.n.@val*0.017453292519943295)
				matP = MatMult(matP,m)
				matV = MatMult(matV,m)
				matN = MatMult(matN,m)
		end
	end
	transform.@matP = matP
	transform.@matV = matV
	transform.@matN = matN
return str

/* --- ReadTitle --- */
ReadTitle:
	line = read(fin)
	fluka.@TITLE = strip(line)
	if outFormat<=@FREE then do
		call lineout fout,FlukaCreateLine(outDigits)
		call lineout fout,line
	end
return

/* --- FixDiscard --- */
FixDiscard:
	do i=1 to 6
		what.i = FlukaParticle(what.i)
	end
	if outFormat<=@FREE then do
		if ^toNumbers then do i=1 to 6
			what.i = FlukaParticleName(what.i)
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixLowMat --- */
FixLowMat:
	w = strip(what.1)
	if ^datatype(w,"NUM") & w<>"" then do
		what.1 = material.w
		if what.i = "?" then call FlError "LOWMAT: Unknown material:" w
	end
	if outFormat<=@FREE then do
		if ^toNumbers then do
			m = Material(what.1); what.1 = material.m
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixUsrTrack --- */
FixUsrTrack:
	if sdum = "&" then do
		if outFormat<=@FREE then
			call lineout fout,FlukaCreateLine(outDigits)
		return
	end
	what.2 = FlukaParticle(what.2)
	what.4 = Region(what.4)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			what.2 = FlukaParticleName(what.2)
			r = what.4; what.4 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixResNuclei --- */
FixResNuclei:
	what.5 = Region(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			r = what.5; what.5 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixDumpThem --- */
FixDumpThem:
	what.5 = Region(what.5)
	what.6 = Region(what.6)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			r = what.5; what.5 = region.r||suffix
			r = what.6; what.6 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixAll --- */
FixAll:
	do ii=1 to 6
		w = strip(what.ii)
		if w<>"" & ^datatype(w,"N") then do
			/* Check for Particle */
			part = FlukaParticle(w)
			if part<>"?" & part<>w then do
				what.ii = part
				iterate
			end
			/* Check for Region */
			reg = region.w
			if reg<>"?" & reg<>w then do
				what.ii = reg
				iterate
			end
			/* Check for Material */
			mat = material.w
			if mat<>"?" then do
				what.ii = mat
				iterate
			end
		end
	end
	if outFormat<=@FREE then
		call lineout fout,FlukaCreateLine(outDigits)
return

/* --- FixPartThr --- */
FixPartThr:
	what.2 = FlukaParticle(what.2)
	what.3 = FlukaParticle(what.3)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			what.2 = FlukaParticleName(what.2)
			what.3 = FlukaParticleName(what.3)
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixScore --- */
FixScore:
	do i=1 to 4
		what.i = FlukaParticle(what.i)
	end
	if outFormat<=@FREE then do
		if ^toNumbers then do i=1 to 4
			what.i = FlukaParticleName(what.i)
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixUsrXXX --- */
FixUsrXXX:
	if sdum = "&" then do
		if outFormat<=@FREE then
			call lineout fout,FlukaCreateLine(outDigits)
		return
	end
	what.2 = FlukaParticle(what.2)
	if outFormat<=@FREE then do
		if ^toNumbers then
			what.2 = FlukaParticleName(what.2)
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixUsrBdx --- */
FixUsrBdx:
	if sdum = "&" then do
		if outFormat<=@FREE then
			call lineout fout,FlukaCreateLine(outDigits)
		return
	end
	what.2 = FlukaParticle(what.2)
	what.4 = Region(what.4)
	what.5 = Region(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			what.2 = FlukaParticleName(what.2)
			r = what.4; what.4 = region.r||suffix
			r = what.5; what.5 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixUsrBin --- */
FixUsrBin:
	if sdum = "&" then do
		if _regbin then do
			what.1 = Region(what.1)
			what.2 = Region(what.2)
			what.3 = Region(what.3)
		end
	end; else do
		call AddUsr "@BIN",sdum
		what.2 = FlukaParticle(what.2)
		if trunc(what.1) = 2 then do
			what.4 = Region(what.4)
			what.5 = Region(what.5)
			what.6 = Region(what.6)
			_regbin = 1
		end; else
			_regbin = 0
	end
	if outFormat<=@FREE then do
		if ^toNumbers & sdum<>"&" then
			what.2 = FlukaParticleName(what.2)
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixTcquench --- */
FixTcquench:
	if outFormat<=@FREE then do
		what.4 = GetUsr("@BIN",what.4)
		what.5 = GetUsr("@BIN",what.5)
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixEmfCut --- */
FixEmfCut:
	if wordpos(sdum,"PROD-CUT ELPO-THR ANNH-THR PHOT-THR PHO2-THR")>0 then
		call FixMat45
	else
		call FixReg45
return

/* --- FixEmfBias --- */
FixEmfBias:
	if wordpos(sdum,"LAMBEMF LAMBCOMP LAMBBREM LBRREMF LBRRCOMP LBRRBREM")>0 then
		call FixMat45
	else
		call FixReg45
return

/* --- FixReg45 --- */
FixReg45:
	what.4 = Region(what.4)
	what.5 = Region(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			r = what.4; what.4 = region.r||suffix
			r = what.5; what.5 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixMat45 --- */
FixMat45:
	what.4 = Material(what.4)
	what.5 = Material(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			m = Material(what.4); what.4 = material.m
			m = Material(what.5); what.5 = material.m
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixLamBias --- */
FixLamBias:
	what.3 = Material(what.3)
	what.4 = FlukaParticle(what.4)
	what.5 = FlukaParticle(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			m = Material(what.3); what.3 = material.m
			what.4 = FlukaParticleName(what.4)
			what.5 = FlukaParticleName(what.5)
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixLowBias --- */
FixLowBias:
	what.4 = Region(what.4)
	what.5 = Region(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			r = what.4; what.4 = region.r||suffix
			r = what.5; what.5 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- ReadGlobal --- */
ReadGlobal:
	if outFormat>@FREE then return
	if outFormat<@FREE then
		what.5 = ""
	else
		what.5 = 1
	call lineout fout,FlukaCreateLine(outDigits)
return

/* --- AddUsr --- */
AddUsr: procedure expose usr.
	parse arg type, sdum
	N = usr.type.0 + 1
	usr.type.0  = N
	usr.type.N = sdum
	usr.type.sdum = N
return

/* --- GetUsr --- */
GetUsr: procedure expose usr. toNumbers
	parse arg type, name
	name = strip(name)
	if name="" then return name
	if datatype(name,"N") && toNumbers then do
		uname = usr.type.name
		if uname="?" then call FLError "USR"||substr(type,2) "name" name "not found!"
		return uname
	end
return name

/* --- Read Material --- */
ReadMaterial:
	material = sdum

	if what.4 = "" then do
		material.0 = material.0 + 1
		what.4 = material.0
	end

	if what.4 <> 0 then do
		m = trunc(what.4)
		material.sdum = m
		material.m = sdum
		if verbose then
			say "Material=" right(m,2) sdum
		material.0 = max(material.0,m)
		material.m.@compound = ""
	end; else do
	end

	material.m.@density = what.3+0
	material.m.@used = 0

	/* Add always a @LASTMAT */
	if m=material.0 then do
		m1 = m + 1;	name = "@LASTMAT"
		material.m1 = name
		material.name = m
	end

	if what.1 <> "" & what.1 <> 0 then do	/* simple material */
		if MAZ.material ^= "" then do
			material.m.@ZAID = word(MAZ.material,1)
			material.m.@compound = MAZ.material
		end; else do
			Z = trunc(format(what.1))
			if what.6 <> "" then
				A = trunc(format(what.6))
			else
				A = trunc(format(what.2))
			material.m.@ZAID = Z*1000 + A
			material.m.@compound = (Z*1000+A) 1.0
		end
	end

	if outFormat<=@FREE then do
		if ^toNumbers then what.4 = ""
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- Material --- */
Material: procedure expose material. fluka.
	parse arg m
	m = strip(m)
	if m="" then return m
	if datatype(m,'N') then return trunc(m)
	neg = left(m,1)=="-"
	if neg | left(m,1)=="+" then m = substr(m,2)
	mat = material.m
	if mat=="?" then call FlError "MATERIAL: Unknown material:" mat m
	if neg then mat = -mat
return mat

/* --- ReadCompound --- */
ReadCompound:
	m = material.sdum
	if m="?" | material.m="?" then call FLError "Material" sdum "is not defined"
	if material.m.@compound = "?" then material.m.@compound = ""
	if material.m.@compound = "0 1" then material.m.@compound = ""

	do fr=1 to 5 by 2
		idx = fr+1
		what.idx = strip(what.idx)
		if what.idx="" | what.idx=0 then leave

		what.idx = Material(what.idx)

		if what.idx>0 then do
			fraction = what.fr	/* atom | mass relative */
			if ^datatype(fraction,"N") then
				call FLError "Not a valid number in compound" fraction
			isotope = trunc(what.idx)
			mc = material.isotope.@compound
			if mc="?" then call FlError "Unknown isotope:" what.idx "in" sdum
			do while mc <> ""
				parse var mc maz f mc
				material.m.@compound = material.m.@compound,
					maz abs(f)*fraction
			end
		end; else
		if what.idx=0 then
			nop
		else do				/* volume fraction */
			fraction = abs(what.fr)
			isotope = abs(trunc(what.idx))
			mc = material.isotope.@compound
			if mc="?" then call FlError "Unknown isotope:" what.idx "in" sdum
			rho = material.isotope.@density
			do while mc <> ""
				parse var mc maz f mc
				material.m.@compound = material.m.@compound,
					maz abs(f)*fraction*rho
			end
			call FlWarning "Compound:" sdum m" Volume fractions might be wrong!"
		end

	end

	if outFormat<=@FREE then do
		if ^toNumbers then do i=2 to 6 by 2
			if what.i<>"" then do
				m = Material(what.i)
				what.i = material.m
			end
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- ReadAssignMat --- */
ReadAssignMat:
	m = trunc(Material(what.1))
	if m="?" | material.m="?" then call FLError "Material" what.1 "is not defined"
	what.1 = m
	what.2 = Region(what.2)
	if what.3<>"" then what.3 = Region(what.3)

	do r=Region(what.2) to Region(what.3,what.2) by Step(what.4)
		region.r.@mat = m
		material.m.@used = 1
		if verbose then
			say "Region" r "material" material.m "("m")"
	end

	if outFormat<=@FREE then do
		if ^toNumbers then do
			what.1 = material.m
			r = what.2;	what.2 = region.r||suffix
			r = what.3; if r<>"" then what.3 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- FixDeltaRay --- */
FixDeltaRay:
	what.4 = Material(what.4)
	what.5 = Material(what.5)

	if outFormat<=@FREE then do
		if ^toNumbers then do
			m = what.4; what.4 = material.m
			m = what.5; if m<>"" then what.5 = material.m
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- ReadBiasing --- */
ReadBiasing:
	importance = what.3
	if what.4 = "" | what.4 = 0 then return
	what.4 = Region(what.4)
	what.5 = Region(what.5)
	do r=Region(what.4) to Region(what.5,what.4) by Step(what.6)
		region.r.@bias = importance
		if verbose then
			say "Region" r "importance" importance
	end
	if outFormat<=@FREE then do
		if ^toNumbers then do
			r = what.4; what.4 = region.r||suffix
			r = what.5; what.5 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end
return

/* --- ReadPlotGeom --- */
ReadPlotGeom:
	if outFormat<=@FREE then do
		call lineout fout,FlukaCreateLine(outDigits)
		do 4
			call lineout fout,read(fin)
		end
	end
return

/* --- Step --- */
Step: if arg(1)="" | arg(1)=0 then return 1; else return arg(1)

/* --- ReadGeometry --- */
ReadGeometry:
	/* --- Check if we use another file for the geometry --- */
	if sdum = "COMBNAME" then geoFormat=@FREE
	if ^datatype(what.3,"N") then what.3 = 5
	if trunc(what.3) <> 5 then do
		line = strip(read(fin))
		if verbose then say "Using file '"line"' for geometry"
		fgeo = open(line,"r")
		if fgeo<0 then do
			call lineout "<STDERR>","ERROR opening fluka geometry file '"line"'"
			exit 4
		end
	end; else
		fgeo = fin

	line = FlukaReadLine(fgeo)

	ivopt = left(line,5)
	idbg  = substr(line,6,5)
	if ivopt="" then ivopt=0
	if idbg=""  then idbg=0
	if geoFormat=0 then do
		if idbg<0  then geoFormat = 1
	end

	select
		when outFormat=@DOUBLE then idbg="-10"
		when outFormat=@FREE then idbg="0"
		otherwise
			idbg="0"
	end

	if verbose then
		say "Geometry Format:",
			word("SINGLE-PRECISSION DOUBLE-PRECISION FREE",geoFormat+1)

	call ReadPrimitives
	call ReadRegions
	call ReadLattices
	call FlukaFindCard fin,"GEOEND"
	if transform.0 > 0 then do
		numeric digits transform.@accuracy
		call TransformPrimitives
	end
return

/* ====================== PRIMITIVES ======================= */
/* --- ReadPrimitives --- */
ReadPrimitives: procedure expose fluka. prim. fgeo verbose geoFormat outFormat,
                        VecXo VecYo VecZo fluka.@LINENO,
                        @MCNP @MCNPX @POVRAY @FREE @DXF
	n = 0
	if verbose then say "Reading primitives"
	do forever
		line = FlukaReadPrimitive(fgeo,geoFormat,0)

		if primitive="END" then leave

		n    = n + 1
		if ^datatype(name,"N") then name = strip(name)

                if outFormat=@FREE | outFormat=@POVRAY | outFormat=@DXF then
                        if datatype(name,"N") then name="BDY"right(n,4,"_")

		if prim.name <> "?" then
			call FlWarning "Duplicate primitive found" name

		prim.n          = name
		prim.name       = n
		prim.n.@prim    = primitive
		prim.n.@comment = fluka.@COMMENT
		prim.n.@line    = line
		prim.n.@lineno  = fluka.@LINENO
		prim.n.@neg     = 0

		if verbose then say n primitive name
		select
			when primitive=="RPP" then call Rectangular
			when primitive=="BOX" then call Box
			when primitive=="SPH" then call Sphere
			when primitive=="RCC" then call Cylinder
			when primitive=="REC" then call EllipticalCylinder
			when primitive=="TRC" then call TruncatedCone
			/**
			when primitive=="ELL" then call Ellipsoid
			when primitive=="WED" then call AngleWedge
			when primitive=="ARB" then call Polyhedron
			**/
			when primitive=="XYP" then call Plane 'Z'
			when primitive=="XZP" then call Plane 'Y'
			when primitive=="YZP" then call Plane 'X'
			when primitive=="PLA" then call Plane 'A'
			when primitive=="XCC" then call InfCylinder 'X'
			when primitive=="YCC" then call InfCylinder 'Y'
			when primitive=="ZCC" then call InfCylinder 'Z'
			when primitive=="XEC" then call InfEllipticalCylinder 'X'
			when primitive=="YEC" then call InfEllipticalCylinder 'Y'
			when primitive=="ZEC" then call InfEllipticalCylinder 'Z'
			otherwise
				call FLError "Unknown body type '"primitive"'"
		end
	end
	prim.0 = n
return

/* --- TransformPrimitives --- */
TransformPrimitives: procedure expose fluka. verbose prim. transform.,
				VecXo VecYo VecZo region. FarAway
	if verbose then say "Transforming Primitives"
	start = transform.@start
	stop  = transform.@stop
	do i=1 to prim.0
		primitive = prim.i.@prim
		if start<>"" then do
			if prim.i<>start then
				iterate
			else
				start=""
		end
		if stop<>"" then if prim.i=stop then leave
		if verbose then say ">>>" right(i,4) prim.i primitive
		select
			when primitive=="RPP" then call TRectangular i
			when primitive=="BOX" then call TBox i
			when primitive=="SPH" then call TSphere i
			when primitive=="RCC" then call TCylinder i
			when primitive=="REC" then call TEllipticalCylinder i
			when primitive=="TRC" then call TTruncatedCone i
			/**
			when primitive=="ELL" then call TEllipsoid i
			when primitive=="WED" then call TAngleWedge
			when primitive=="ARB" then call TPolyhedron
			**/
			when primitive=="XYP" | primitive=="XZP" |,
			     primitive=="YZP" | primitive=="PLA" then call TPlane i
			when primitive=="XCC" | primitive=="YCC" |,
			     primitive=="ZCC" then call TInfCylinder i
			when primitive=="XEC" | primitive=="YEC" |,
			     primitive=="ZEC" then call TInfEllipticalCylinder i
			otherwise
				call FLError "Transform Unknown body type" primitive
		end
	end

        /* Correct regions description for NEGative bodies */
	do r=1 to region.0
		/* Scan over products */
		do i=1 to region.r.0
			product = region.r.i
			newproduct = ""
			changed = 0
			do while product<>""
				parse var product p product
				parse var p sign +1 pr
				if sign<>"-" & sign<>"+" then
					call FlError "Region" r region.r "Sign missing" p
				if ^datatype(pr,"N") then
					pn = prim.pr
				else
					pn = pr
				if pr = "(" then call FlError "Parenthesis found with parenthesis expansion disabled"
				if prim.pn.@NEG then do /* change the sign */
					if sign=="+" then
						p="-"pr
					else
						p="+"pr
					if verbose then
						say "Changed sign region="r region.r p sign pr
					changed = 1
				end
				newproduct = newproduct p
			end
			if changed then
				region.r.i = strip(newproduct)
		end
	end
return

/* --- WritePrimitives --- */
WritePrimitives: procedure expose verbose fluka. prim. toNumbers outDigits,
			outFormat @FREE suffix
	parse arg fout
	do n=1 to prim.0
		call WriteMultiLine fout,prim.n.@comment
		if verbose then
			call WriteMultiLine fout,prim.n.@line,"*"

		if toNumbers then
			name = right(n,4)
		else
			name = prim.n

		if outFormat=2 then
			if ^datatype(name,"N") then name=name||suffix

		primitive = prim.n.@prim
		sdum = ""
		what. = ""

		select
			when primitive=="RPP" then call WriteRectangular
			when primitive=="BOX" then call WriteBox
			when primitive=="SPH" then call WriteSphere
			when primitive=="RCC" then call WriteCylinder
			when primitive=="REC" then call WriteEllipticalCylinder
			when primitive=="TRC" then call WriteTruncatedCone
			/**
			when primitive=="ELL" then call WriteEllipsoid
			when primitive=="WED" then call WriteAngleWedge
			when primitive=="ARB" then call WritePolyhedron
			**/
			when primitive=="XYP" then call WritePlane 'Z'
			when primitive=="XZP" then call WritePlane 'Y'
			when primitive=="YZP" then call WritePlane 'X'
			when primitive=="PLA" then call WritePlane 'A'
			when primitive=="XCC" then call WriteInfCylinder 'X'
			when primitive=="YCC" then call WriteInfCylinder 'Y'
			when primitive=="ZCC" then call WriteInfCylinder 'Z'
			when primitive=="XEC" then call WriteInfEllipticalCylinder 'X'
			when primitive=="YEC" then call WriteInfEllipticalCylinder 'Y'
			when primitive=="ZEC" then call WriteInfEllipticalCylinder 'Z'
		end
	end
	call lineout fout,"  END"
return

/* --- WhatNeeded --- */
WhatNeeded:
	do while what.0<arg(1)
		prim.n.@line = prim.n.@line"0A"x||,
			FlukaReadPrimitive(fgeo,geoFormat,1)
	end
return

/* --- Rectangular --- */
Rectangular:
	call WhatNeeded 6
	prim.n.@xmin = what.1
	prim.n.@xmax = what.2
	prim.n.@ymin = what.3
	prim.n.@ymax = what.4
	prim.n.@zmin = what.5
	prim.n.@zmax = what.6
return

WriteRectangular:
	what.0 = 6
	what.1 = prim.n.@xmin
	what.2 = prim.n.@xmax
	what.3 = prim.n.@ymin
	what.4 = prim.n.@ymax
	what.5 = prim.n.@zmin
	what.6 = prim.n.@zmax
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TRectangular: procedure expose prim. transform. VecXo VecYo VecZo verbose
	parse arg n

	prim.n.@prim = "BOX"		/* Convert to box */
	prim.n.@P = VecMake(prim.n.@xmin, prim.n.@ymin, prim.n.@zmin)
	prim.n.@X = VecMake(prim.n.@xmax-prim.n.@xmin, 0.0, 0.0)
	prim.n.@Y = VecMake(0.0, prim.n.@ymax-prim.n.@ymin, 0.0)
	prim.n.@Z = VecMake(0.0, 0.0, prim.n.@zmax-prim.n.@zmin)
	call TBox n
	if prim.n.@prim = "BOX" & verbose then
		say "Converting RPP to BOX" prim.n
return

/* --- Box --- */
Box:
	call WhatNeeded 12
	prim.n.@P = VecMake(what.1,  what.2,  what.3)
	prim.n.@X = VecMake(what.4,  what.5,  what.6)
	prim.n.@Y = VecMake(what.7,  what.8,  what.9)
	prim.n.@Z = VecMake(what.10, what.11, what.12)

	/* Correct the order */
	if VecDot(VecCross(prim.n.@X,prim.n.@Y),prim.n.@Z)<0 then do
		/* inverse X with Y */
		tmp       = prim.n.@X
		prim.n.@X = prim.n.@Y
		prim.n.@Y = tmp
	end
return

WriteBox:
	what.0  = 12
	what.1  = VecX(prim.n.@P)
	what.2  = VecY(prim.n.@P)
	what.3  = VecZ(prim.n.@P)
	what.4  = VecX(prim.n.@X)
	what.5  = VecY(prim.n.@X)
	what.6  = VecZ(prim.n.@X)
	what.7  = VecX(prim.n.@Y)
	what.8  = VecY(prim.n.@Y)
	what.9  = VecZ(prim.n.@Y)
	what.10 = VecX(prim.n.@Z)
	what.11 = VecY(prim.n.@Z)
	what.12 = VecZ(prim.n.@Z)
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TBox: procedure expose prim. transform. VecXo VecYo VecZo verbose
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@X = MatXVec(transform.@matV, prim.n.@X)
	prim.n.@Y = MatXVec(transform.@matV, prim.n.@Y)
	prim.n.@Z = MatXVec(transform.@matV, prim.n.@Z)

	/* TODO Optimize Box (convert to RPP if necessary) */
	/* Check X,Y,Z against Xo, Yo, Zo... */
	dX = right(CheckDirection(prim.n.@X),1)
	dY = right(CheckDirection(prim.n.@Y),1)
	dZ = right(CheckDirection(prim.n.@Z),1)
	if dX<>"O" & dY<>"O" & dZ<>"O" then do
		if verbose then say "Converting BOX to RPP" prim.n

		/* Convert to box */
		vX = FindVecDir("X")
		vY = FindVecDir("Y")
		vZ = FindVecDir("Z")

		prim.n.@prim = "RPP"
		prim.n.@xmin = MIN(VecX(prim.n.@P),VecX(prim.n.@P)+vX)
		prim.n.@xmax = MAX(VecX(prim.n.@P),VecX(prim.n.@P)+vX)
		prim.n.@ymin = MIN(VecY(prim.n.@P),VecY(prim.n.@P)+vY)
		prim.n.@ymax = MAX(VecY(prim.n.@P),VecY(prim.n.@P)+vY)
		prim.n.@zmin = MIN(VecZ(prim.n.@P),VecZ(prim.n.@P)+vZ)
		prim.n.@zmax = MAX(VecZ(prim.n.@P),VecZ(prim.n.@P)+vZ)
	end
return

FindVecDir:
	index = pos(arg(1),"XYZ")
	select
		when dX=arg(1) then return word(prim.n.@X,index)
		when dY=arg(1) then return word(prim.n.@Y,index)
		when dZ=arg(1) then return word(prim.n.@Z,index)
	end
return

/* --- Sphere --- */
Sphere:
	call WhatNeeded 4
	prim.n.@P = VecMake(what.1,what.2,what.3)
	prim.n.@R = what.4
return

WriteSphere:
	what.0 = 4
	what.1 = VecX(prim.n.@P)
	what.2 = VecY(prim.n.@P)
	what.3 = VecZ(prim.n.@P)
	what.4 = prim.n.@R
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TSphere: procedure expose prim. transform.
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@R = prim.n.@R * transform.@scaleR
return

/* --- Cylinder --- */
Cylinder:
	call WhatNeeded 7
	prim.n.@P = VecMake(what.1, what.2, what.3)
	prim.n.@Z = VecMake(what.4, what.5, what.6)
	prim.n.@R = what.7
return

WriteCylinder:
	what.0 = 7
	what.1 = VecX(prim.n.@P)
	what.2 = VecY(prim.n.@P)
	what.3 = VecZ(prim.n.@P)
	what.4 = VecX(prim.n.@Z)
	what.5 = VecY(prim.n.@Z)
	what.6 = VecZ(prim.n.@Z)
	what.7 = prim.n.@R
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TCylinder: procedure expose prim. transform.
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@Z = MatXVec(transform.@matV, prim.n.@Z)
	prim.n.@R = prim.n.@R * transform.@scaleR
return

/* --- InfCylinder --- */
InfCylinder:
	parse arg axis

	call WhatNeeded 3

	prim.n.@R = what.3
	select
		when axis=='X' then do
			prim.n.@P = VecMake(0.0, what.1, what.2)
			prim.n.@N = VecXo
		end
		when axis=='Y' then do
			prim.n.@P = VecMake(what.2, 0.0, what.1)
			prim.n.@N = VecYo
		end
		when axis=='Z' then do
			prim.n.@P = VecMake(what.1, what.2, 0.0)
			prim.n.@N = VecZo
		end
	end
return

WriteInfCylinder:
	parse arg axis
	what.0 = 3
	select
		when axis=='X' then do
			what.1 = VecY(prim.n.@P)
			what.2 = VecZ(prim.n.@P)
		end
		when axis=='Y' then do
			what.1 = VecZ(prim.n.@P)
			what.2 = VecX(prim.n.@P)
		end
		when axis=='Z' then do
			what.1 = VecX(prim.n.@P)
			what.2 = VecY(prim.n.@P)
		end
	end
	what.3 = prim.n.@R
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TInfCylinder: procedure expose prim. transform. VecXo VecYo VecZo FarAway
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@N = MatXVec(transform.@matN, prim.n.@N)
	prim.n.@R = prim.n.@R * transform.@scaleR

	dir = right(CheckDirection(prim.n.@N),1)
	select
		when dir=="X" then prim.n.@prim = "XCC"
		when dir=="Y" then prim.n.@prim = "YCC"
		when dir=="Z" then prim.n.@prim = "ZCC"
		when dir=="O" then do
			call FlWarning "Converting an Infinite Cylinder [XYZ]CC to RCC"
                        prim.n.@prim = "RCC"
                        prim.n.@P = VecAddS(-FarAway,prim.n.@N,prim.n.@P)
                        prim.n.@Z = VecS(2*FarAway,prim.n.@N)
                end
	end
return

/* --- InfEllipticalCylinder --- */
InfEllipticalCylinder:
	parse arg axis

	call WhatNeeded 4
	select
		when axis=='X' then do
			prim.n.@P = VecMake(0.0, what.1, what.2)
			prim.n.@R = VecMake(0.0, what.3, what.4)
			prim.n.@X = VecMake(0.0, what.3, 0.0)
			prim.n.@Y = VecMake(0.0, 0.0,    what.4)
			prim.n.@N = VecXo
		end
		when axis=='Y' then do
			prim.n.@P = VecMake(what.2, 0.0, what.1)
			prim.n.@R = VecMake(what.4, 0.0, what.3)
			prim.n.@X = VecMake(0.0,    0.0, what.3)
			prim.n.@Y = VecMake(what.4, 0.0, 0.0)
			prim.n.@N = VecYo
		end
		when axis=='Z' then do
			prim.n.@P = VecMake(what.1, what.2, 0.0)
			prim.n.@R = VecMake(what.3, what.4, 0.0)
			prim.n.@X = VecMake(what.3, 0.0,    0.0)
			prim.n.@Y = VecMake(0.0,    what.4, 0.0)
			prim.n.@N = VecZo
		end
	end
return

WriteInfEllipticalCylinder:
	parse arg axis
	what.0 = 4
	select
		when axis=='X' then do
			what.1 = VecY(prim.n.@P)
			what.2 = VecZ(prim.n.@P)
			what.3 = VecY(prim.n.@R)
			what.4 = VecZ(prim.n.@R)
		end
		when axis=='Y' then do
			what.1 = VecZ(prim.n.@P)
			what.2 = VecX(prim.n.@P)
			what.3 = VecZ(prim.n.@R)
			what.4 = VecX(prim.n.@R)
		end
		when axis=='Z' then do
			what.1 = VecX(prim.n.@P)
			what.2 = VecY(prim.n.@P)
			what.3 = VecX(prim.n.@R)
			what.4 = VecY(prim.n.@R)
		end
	end
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TInfEllipticalCylinder: procedure expose prim. transform. VecXo VecYo VecZo
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@R = MatXVec(transform.@matV, prim.n.@R)
	prim.n.@X = MatXVec(transform.@matV, prim.n.@X)
	prim.n.@Y = MatXVec(transform.@matV, prim.n.@Y)

	dir = right(CheckDirection(prim.n.@N),1)
	select
		when dir=="X" then prim.n.@prim = "XEC"
		when dir=="Y" then prim.n.@prim = "YEC"
		when dir=="Z" then prim.n.@prim = "ZEC"
		when dir=="O" then do
			call FlWarning "Converting an Infinite Elliptical Cylinder [XYZ]EC to REC"
                        prim.n.@prim = "REC"
                        prim.n.@P = VecAddS(-FarAway,prim.n.@N,prim.n.@P)
                        prim.n.@Z = VecS(2*FarAway,prim.n.@N)
		end
	end
return

/* --- EllipticalCylinder --- */
EllipticalCylinder:
	call WhatNeeded 12
	prim.n.@P = VecMake(what.1,  what.2,  what.3)
	prim.n.@Z = VecMake(what.4,  what.5,  what.6)
	prim.n.@X = VecMake(what.7,  what.8,  what.9)
	prim.n.@Y = VecMake(what.10, what.11, what.12)
	/* Correct the order */
	if VecDot(VecCross(prim.n.@X,prim.n.@Y),prim.n.@Z)<0 then do
		parse var prim.n.@X x y z
		prim.n.@X = VecMake(-x,-y,-z)
	end
return

WriteEllipticalCylinder:
	what.0  = 12
	what.1  = VecX(prim.n.@P)
	what.2  = VecY(prim.n.@P)
	what.3  = VecZ(prim.n.@P)
	what.4  = VecX(prim.n.@Z)
	what.5  = VecY(prim.n.@Z)
	what.6  = VecZ(prim.n.@Z)
	what.7  = VecX(prim.n.@X)
	what.8  = VecY(prim.n.@X)
	what.9  = VecZ(prim.n.@X)
	what.10 = VecX(prim.n.@Y)
	what.11 = VecY(prim.n.@Y)
	what.12 = VecZ(prim.n.@Y)
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TEllipticalCylinder: procedure expose prim. transform.
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@X = MatXVec(transform.@matV, prim.n.@X)
	prim.n.@Y = MatXVec(transform.@matV, prim.n.@Y)
	prim.n.@Z = MatXVec(transform.@matV, prim.n.@Z)
return

/* --- TruncatedCone --- */
TruncatedCone:
	call WhatNeeded 8
	prim.n.@P = VecMake(what.1, what.2, what.3)
	prim.n.@Z = VecMake(what.4, what.5, what.6)
	prim.n.@bR = what.7
	prim.n.@aR = what.8

return

TTruncatedCone: procedure expose prim. transform.
	parse arg n
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@Z = MatXVec(transform.@matV, prim.n.@Z)
	prim.n.@bR = prim.n.@bR * transform.@scaleR
	prim.n.@aR = prim.n.@aR * transform.@scaleR
return

WriteTruncatedCone:
	what.0  = 8
	what.1  = VecX(prim.n.@P)
	what.2  = VecY(prim.n.@P)
	what.3  = VecZ(prim.n.@P)
	what.4  = VecX(prim.n.@Z)
	what.5  = VecY(prim.n.@Z)
	what.6  = VecZ(prim.n.@Z)
	what.7  = prim.n.@bR
	what.8  = prim.n.@aR
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

/* --- Plane --- */
Plane:
	parse arg axis
	select
		when axis=='X' then do
			prim.n.@P = VecMake(what.1, 0.0, 0.0)
			prim.n.@N = VecXo
			prim.n.@PP = what.1
		end

		when axis=='Y' then do
			prim.n.@P = VecMake(0.0, what.1, 0.0)
			prim.n.@N = VecYo
			prim.n.@PP = what.1
		end

		when axis=='Z' then do
			prim.n.@P = VecMake(0.0, 0.0, what.1)
			prim.n.@N = VecZo
			prim.n.@PP = what.1
		end

		otherwise
			call WhatNeeded 6
			prim.n.@P = VecMake(what.4, what.5, what.6)
			prim.n.@N = VecMake(what.1, what.2, what.3)
	end
return

WritePlane:
	parse arg axis
	what.0 = 1
	select
		when axis=='X' then what.1 = VecX(prim.n.@P)
		when axis=='Y' then what.1 = VecY(prim.n.@P)
		when axis=='Z' then what.1 = VecZ(prim.n.@P)
		otherwise
			what.0 = 6
			what.1 = VecX(prim.n.@N)
			what.2 = VecY(prim.n.@N)
			what.3 = VecZ(prim.n.@N)
			what.4 = VecX(prim.n.@P)
			what.5 = VecY(prim.n.@P)
			what.6 = VecZ(prim.n.@P)
	end
	call lineout fout, FlukaCreatePrimitiveLine(outFormat)
return

TPlane: procedure expose prim. transform. VecXo VecYo VecZo
	parse arg n
	prim.n.@prim = "PLA"
	prim.n.@P = MatXVec(transform.@matP, prim.n.@P)
	prim.n.@N = MatXVec(transform.@matN, prim.n.@N)
	dir = CheckDirection(prim.n.@N)
	d = right(dir,1)
	select
		when d=="X" then prim.n.@prim = "YZP"
		when d=="Y" then prim.n.@prim = "XZP"
		when d=="Z" then prim.n.@prim = "XYP"
	end
	if left(dir,1)=="-" then prim.n.@NEG = 1
return

/* ====================== REGIONS ======================= */
/* --- WriteRegions --- */
WriteRegions: procedure expose verbose fluka. region. prim. toNumbers,
			outFormat @FREE suffix
	parse arg fout

	do r=1 to region.0
		call WriteMultiLine fout,region.r.@comment
		fluka.@LINENO = region.r.@lineno
		fluka.@LINE   = region.r.@line

		if verbose then
			call WriteMultiLine fout,region.r.@line,"*"

		if outFormat<>@FREE then do
			ors = "OR"
			nspace = copies(" ",10)
		end; else do
			ors = "| "
			nspace = copies(" ",14)
		end

		/* Scan over products */
		do i=1 to region.r.0
			if toNumbers then
				reg = right(r,3,0)
			else do
				reg = region.r
				if outFormat=@FREE then reg=reg||suffix
			end

			if i=1	then do
				if region.r.@n = "" then region.r.@n = 5
				if outFormat<>@FREE then
					line = right(reg,5)||right(region.r.@n,5)
				else
					line = left(reg,10)||right(region.r.@n,3)" "

			end; else
				line = nspace

			product = region.r.i
			first = 1
			do while product<>""
				parse var product p product

				neg = left(p,1)=="-"
				if neg | left(p,1)=="+" then p = substr(p,2)

				if datatype(p,"N") && toNumbers then do
					if prim.p == "?" then
						call FlError "Unknown primitive: '"p"'"
					/*if datatype(p,"N") then
						if prim.p.@neg then neg=^neg */
					p = prim.p
					/*if datatype(p,"N") then
						if prim.p.@neg then neg=^neg*/
				end

				if neg	then p = "-"p
					else p = "+"p

				if first then do
					if region.r.0>1 then line = line||ors
							else line = line"  "
					first = 0
				end; else
					line = line"  "

				if outFormat<>@FREE then
					line = line || right(p,5)
				else
					line = line || left(p||suffix,10)

				if length(line)>72 then do
					call lineout fout,strip(line,"T")
					line = nspace
				end
			end
			if line<>"" then
				call lineout fout,strip(line,"T")
		end
	end
	call lineout fout,"  END"
return

/* --- ReadRegions --- */
ReadRegions: procedure expose fluka. prim. region. fgeo verbose blackbody,
			geoFormat outFormat optimizeGeom fluka.@LINENO,
                        @MCNP @MCNPX @POVRAY @FREE @DXF
	r = 0
	if verbose then say "Reading regions"
	do forever
		line = FlukaReadRegion(fgeo,geoFormat)
		if region="END" then leave

		r = r + 1
                if outFormat=@FREE | outFormat=@POVRAY | outFormat=@DXF then
                        if datatype(region,"N") then region="RGN"right(r,4,"_")

		if region.region <> "?" then
			call FlWarning "Duplicate region found" region

		region.region     = r
		region.r          = region
		region.r.@line    = line
		region.r.@lineno  = fluka.@LINENO
		region.r.@comment = fluka.@COMMENT
		region.r.@expr    = region_expr
		region.r.@n       = region_n
		if verbose then say r region

		if geoFormat=2 & ^datatype(region_n,"N") then
			call FLError "Region" region "is not followed by a number"

		ors = 0
		if region_expr="" then
			call FlError "Region" r region "resulted in any empty definition!"

		do while region_expr<>""
			parse var region_expr product "|" region_expr
			if product = "" then iterate
			if optimizeGeom then
				product = OptimizeRegionProduct(product)
			else do
				product = changestr("+",product," +")
				product = changestr("-",product," -")
				product = strip(product)
			end
			if product ^= "" then do
				ors = ors + 1
				region.r.ors = product
			end
		end
		region.r.0 = ors

		if verbose then
			do i=1 to region.r.0
				say "	OR" region.r.i
			end
	end

	if outFormat=@MCNP | outFormat=@MCNPX then do
		/* Add universe */
		r = r + 1
		region = "Universe"
		region.region = "r"
		region.r = region
		region.@comment = "Universe"
		region.r.@line  = "Universe"
		region.r.0 = 1
		region.r.1 = "-"blackbody
		region.r.@mat = 1
	end

	region.0 = r
	r = r+1;	name="@LASTREG"
	region.r = name
	region.name = r-1
return

/* --- Region --- */
Region: procedure expose region. fluka. suffix
	r = strip(arg(1))
	if r="" | r=0 then do
		r = strip(arg(2))
		if r="" | r=0 then return r
	end
	if r="" then return r
	if datatype(r,'N') then do
		if r<0 | r>region.0 then r=region.0+1
		return trunc(r)
	end
	if region.r == "?" then call FlError "Unknown region: '"r"'"
return region.r

/* --- OptimizeRegionProduct --- */
OptimizeRegionProduct: procedure expose prim.
	product = arg(1)

	do mainloop=1
		wproduct = translate(product,"  ","+-")
		do a=1 to words(wproduct)
			pa = word(wproduct,a)
			n = prim.pa
			ta = prim.n.@prim
			if ta=="XYP" | ta="XZP" | ta="YZP" then
				if OptimizePlanes() then iterate mainloop
		end
		leave mainloop
	end mainloop
	product = changestr("+",product," +")
	product = changestr("-",product," -")
return strip(product)

/* --- OptimizePlanes --- */
OptimizePlanes:
	do b=a+1 to words(wproduct)
		pb = word(wproduct,b)
		m = prim.pb
		tb = prim.m.@prim
		if tb<>ta then iterate

		/* Find sign */
		sa = substr(product,wordindex(wproduct,a)-1,1)
		sb = substr(product,wordindex(wproduct,b)-1,1)

		/* Find position */
		xa = prim.n.@PP
		xb = prim.m.@PP

		/* Find low and high value */
		del = 0

		if xa<xb then do
			low  = a
			high = b
		end; else
		if xa>xb then do
			low  = b
			high = a
		end; else
			del = b

		if del=0 then
		select
			when sa=="+" & sb=="+" then del=high
			when sa=="-" & sb=="-" then del=low
			when sa=="+" & xa<xb then do
				product = ""
			end
			when sb=="+" & xa>xb then do
				product = ""
			end
		end

		if del>0 then do
			pos = wordindex(wproduct,del)-1
			len = length(word(wproduct,del))+1
			product = space(delstr(product,pos,len),0)
			return 1
		end
	end
return 0

/* --- ReadLattices --- */
ReadLattices: procedure expose fgeo inpFormat outFormat toNumbers lattice. outDigits fluka. region.
	n = 0
	do forever
		line = FlukaReadLine(fgeo)
		if line="" then leave
                if eof(fgeo) then leave
		call FlukaParseLine line,inpFormat
		if card=="GEOEND" then leave
		do i=1 to 2
			what.i = Region(what.i)
			if ^toNumbers then do
				r = what.i
				what.i = region.r||suffix
			end
		end
		n = n + 1
		lattice.n = FlukaCreateLine(outDigits)
		lattice.n.@comment = fluka.@COMMENT
	end
	lattice.0 = n
return

/* --- WriteLattices --- */
WriteLattices: procedure expose lattice. fluka.
	parse arg fout

	do i=1 to lattice.0
		call WriteMultiLine fout,lattice.i.@comment
		call lineout fout,lattice.i
	end
return
	if sdum = "&" then do
		if outFormat<=@FREE then
			call lineout fout,FlukaCreateLine(outDigits)
		return
	end
	what.2 = FlukaParticle(what.2)
	what.4 = Region(what.4)
	what.5 = Region(what.5)
	if outFormat<=@FREE then do
		if ^toNumbers then do
			what.2 = FlukaParticleName(what.2)
			r = what.4; what.4 = region.r||suffix
			r = what.5; what.5 = region.r||suffix
		end
		call lineout fout,FlukaCreateLine(outDigits)
	end

/* ---- LoadMaterialAZ ---- */
LoadMaterialAZ: procedure expose MAZ.
	parse arg filename
	MAZ. = ""
	fin = open(filename,"r")
	if fin<0 then return 0
	do forever
		line = read(fin)
		if eof(fin) then leave
		if left(line,1)="*" then iterate
		parse var line name density az
		if words(az)=1 then az=az "1.0"
		MAZ.name = az
	end
	call close fin
return 1

/* --- WriteMultiLine --- */
WriteMultiLine: procedure
	parse arg fout, multiline, prefix
	do while multiline <> ""
		parse var multiline line "0A"x multiline
		if line <> "" then
			call lineout fout,prefix||line
	end
return

/* --- CheckDirection --- */
CheckDirection: procedure expose transform. VecXo VecYo VecZo
	trace o
	dir = VecNorm(arg(1))
	select
		when VecDot(dir,VecXo) =  1 then return  "X"
		when VecDot(dir,VecYo) =  1 then return  "Y"
		when VecDot(dir,VecZo) =  1 then return  "Z"
		when VecDot(dir,VecXo) = -1 then return "-X"
		when VecDot(dir,VecYo) = -1 then return "-Y"
		when VecDot(dir,VecZo) = -1 then return "-Z"
	end
return "O"

/* ========================== ERROR ========================== */
FLWarning:
	call lineout "<STDERR>","Warning:" arg(1)
	call lineout "<STDERR>","Line:" fluka.@LINENO":" fluka.@LINE
	return

FLError:
	call lineout "<STDERR>","Error:" arg(1)
	call lineout "<STDERR>","Line:" fluka.@LINENO":" fluka.@LINE
	exit 999

Error:
	call lineout "<STDERR>","Error:" RC "at line" SIGL
	call lineout "<STDERR>","Line:" fluka.@LINENO":" fluka.@LINE
	exit 9

Syntax:
	call lineout "<STDERR>","Syntax error:" RC "at line" SIGL
	call lineout "<STDERR>","Line:" fluka.@LINENO":" fluka.@LINE
	exit 9
